(window["webpackJsonp"]=window["webpackJsonp"]||[]).push([["chunk-8b77bb46"],{"006b":function(t,e,r){"use strict";r.r(e);var n=function(){var t=this,e=t.$createElement,r=t._self._c||e;return r(t.rootComponent,{ref:"root",tag:"component",attrs:{template:t.contentTemplate}})},o=[],i=r("2909"),s=r("5530"),a=(r("2ca0"),r("159b"),function(t,e,r){if(!e.hasOwnProperty(r)){var n=Object.getOwnPropertyDescriptor(t,r);Object.defineProperty(e,r,n)}}),c={props:{template:String,parent:Object,templateProps:{type:Object,default:function(){return{}}}},render:function(t){if(this.template){var e=this.parent||this.$parent,r=e.$data;void 0===r&&(r={});var n=e.$props;void 0===n&&(n={});var o=e.$options;void 0===o&&(o={});var i=o.components;void 0===i&&(i={});var s=o.computed;void 0===s&&(s={});var c=o.methods;void 0===c&&(c={});var p=this.$data;void 0===p&&(p={});var u=this.$props;void 0===u&&(u={});var v=this.$options;void 0===v&&(v={});var f=v.methods;void 0===f&&(f={});var h=v.computed;void 0===h&&(h={});var d=v.components;void 0===d&&(d={});var m={$data:{},$props:{},$options:{},components:{},computed:{},methods:{}};Object.keys(r).forEach((function(t){void 0===p[t]&&(m.$data[t]=r[t])})),Object.keys(n).forEach((function(t){void 0===u[t]&&(m.$props[t]=n[t])})),Object.keys(c).forEach((function(t){void 0===f[t]&&(m.methods[t]=c[t])})),Object.keys(s).forEach((function(t){void 0===h[t]&&(m.computed[t]=s[t])})),Object.keys(i).forEach((function(t){void 0===d[t]&&(m.components[t]=i[t])}));var l=Object.keys(m.methods||{}),b=Object.keys(m.$data||{}),g=Object.keys(m.$props||{}),O=Object.keys(this.templateProps),y=b.concat(g).concat(l).concat(O),$=(w=e,P={},l.forEach((function(t){return a(w,P,t)})),P),j=function(t){var e={};return t.forEach((function(t){t&&Object.getOwnPropertyNames(t).forEach((function(r){return a(t,e,r)}))})),e}([m.$data,m.$props,$,this.templateProps]);return t({template:this.template||"<div></div>",props:y,computed:m.computed,components:m.components,provide:this.$parent._provided},{props:j})}var w,P}},p=r("2f62"),u=r("3800"),v={name:"App",components:{VRuntimeTemplate:c},computed:Object(s["a"])(Object(s["a"])(Object(s["a"])({},Object(u["b"])()),Object(p["c"])({get:"APP_GET",contentTemplate:"APP_TEMPLATE",busy:"WS_BUSY",wsClient:"WS_CLIENT",actions:"APP_ACTIONS",routes:"APP_ROUTES"})),{},{rootComponent:function(){return this.contentTemplate.startsWith("#")?this.contentTemplate.substring(1):"v-runtime-template"}}),watch:{actions:function(t){for(var e=0;e<t.length;e++){var r=t[e],n=r.ref,o=r.type;if("method"===o){var s,a=r.method,c=r.args;(s=this.getRef(n))[a].apply(s,Object(i["a"])(c))}if("property"===o){var p=r.property,u=r.value;this.getRef(n)[p]=u}}this.clearActions()},contentTemplate:function(){this.vApp=this.$el},busy:function(t){this.reactiveBusy=t}},data:function(){return{window:window,vApp:null,reactiveBusy:!1}},methods:Object(s["a"])(Object(s["a"])({},Object(p["b"])({serverSet:"APP_SET",serverTrigger:"APP_TRIGGER",clearActions:"APP_ACTIONS_PROCESSED"})),{},{set:function(t,e){this.serverSet({key:t,value:e})},trigger:function(t,e,r){this.serverTrigger({name:t,args:e,kwargs:r})},getRef:function(t){var e=this.$refs.root;return e.$refs&&e.$refs[t]?e.$refs[t]:e.$children[0].$refs[t]}}),provide:function(){var t=this;return{get:function(e){return t.get(e)},set:function(e,r){return t.set(e,r)},trigger:function(e,r,n){return t.trigger(e,r,n)},wsClient:this.wsClient,isBusy:function(){return t.busy},reactiveBusy:this.reactiveBusy,vApp:this.vApp,window:this.window}},mounted:function(){var t=this;this.vApp=this.$el,this.routes.forEach((function(e){t.$router.addRoute(e)}))}},f=v,h=r("2877"),d=Object(h["a"])(f,n,o,!1,null,null,null);e["default"]=d.exports},"2ca0":function(t,e,r){"use strict";var n=r("23e7"),o=r("06cf").f,i=r("50c4"),s=r("5a34"),a=r("1d80"),c=r("ab13"),p=r("c430"),u="".startsWith,v=Math.min,f=c("startsWith"),h=!p&&!f&&!!function(){var t=o(String.prototype,"startsWith");return t&&!t.writable}();n({target:"String",proto:!0,forced:!h&&!f},{startsWith:function(t){var e=String(a(this));s(t);var r=i(v(arguments.length>1?arguments[1]:void 0,e.length)),n=String(t);return u?u.call(e,n,r):e.slice(r,r+n.length)===n}})}}]);
//# sourceMappingURL=chunk-8b77bb46.c6ca42ea.js.map